import requests
from flask import request, redirect
from furl import furl
from .base import Strategy

ACCESS_URL = "https://slack.com/api/oauth.v2.access"
AUTHORIZE_URL = "https://slack.com/oauth/v2/authorize"
PROTO_HEADER = "X-Forwarded-Proto"

class SlackStrategy(Strategy):
    def __init__(self, client_id, client_secret, *, scope):
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = scope

    def _get_current_url(self):
        parsed = furl(request.url)
        # Deals with the scenario that the protocol is HTTP
        # due to termination of the HTTPS request at the load balancer.
        if PROTO_HEADER in request.headers:
            parsed.scheme = request.headers[PROTO_HEADER]
        return parsed.url

    def _get_authorize_url(self):
        parsed = furl(AUTHORIZE_URL)
        parsed.args = dict(
            client_id = self._client_id,
            scope = self._scope,
            redirect_uri = self._get_current_url()
        )
        return parsed.url

    def _get_access_data(self, code):
        parsed = furl(self._get_current_url())
        parsed.args = dict()
        resp = requests.post(ACCESS_URL, data = dict(
            code = code,
            client_id = self._client_id,
            client_secret = self._client_secret,
            redirect_uri = parsed.url
        ))
        if resp.status_code == 200:
            data = resp.json()
            if data["ok"]:
                return data

    # Default behaviour is to just return the access token payload
    # as the "user".
    def get_user(self, data):
        return data

    def authenticate(self, state):
        state.attempted = True
        if "code" not in request.args:
            return redirect(self._get_authorize_url())
        data = self._get_access_data(request.args["code"])
        if data is None:
            return 
        state.user = self.get_user(data)


from datetime import datetime
from .cfg import Configuration
from flask import request
from jwt.exceptions import InvalidTokenError
import jwt
from .base import Strategy, get_user

class JWTStrategy(Strategy):
    def __init__(self, secret, *, cookie = Configuration.default_jwt_cookie, expiry = None):
        self._secret = secret
        self._cookie = cookie
        self._expiry = expiry

    def authenticate(self, state):
        if self._cookie not in request.cookies:
            return
        state.attempted = True
        token = request.cookies[self._cookie]
        try:
            data = jwt.decode(token, self._secret)
            state.user = self.get_user(data)
        except InvalidTokenError:
            pass

    def grant(self, request):
        data = self.get_token(get_user())
        if self._expiry is not None:
            data["exp"] = datetime.utcnow() + self._expiry
        token = jwt.encode(data, self._secret)
        request.set_cookie(self._cookie, token)


from functools import wraps
from flask import redirect, g
from .error import AuthenticationError

AUTH_STATE_KEY = "_lonny_flask_auth"

class AuthState:
    def __init__(self):
        self.attempted = False
        self.user = None

def get_auth_state():
    return g.setdefault(AUTH_STATE_KEY, AuthState())

def get_user():
    return get_auth_state().user

class Middleware:
    def __call__(self, fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            next_fn = lambda: fn(*args, **kwargs)
            return self.run(next_fn)
        return wrapped

class Strategy(Middleware):
    def run(self, next_fn):
        state = get_auth_state()
        if not state.attempted:
            response = self.authenticate(state)
            if response is not None:
                return response
        return next_fn()

class Guard(Middleware):
    def run(self, next_fn):
        if get_auth_state().user is None:
            raise AuthenticationError()
        return next_fn()


from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name = "lonny_flask_auth",
    version = "1.1",
    packages = find_packages(),
    install_requires = requirements
)


from .base import Strategy, Guard, get_user
from .jwt import JWTStrategy
from .api_key import APIKeyStrategy
from .slack import SlackStrategy
from .cfg import Configuration
from .error import AuthenticationError

__all__ = [
    get_user,
    Guard,
    Strategy,
    JWTStrategy,
    APIKeyStrategy,
    SlackStrategy,
    Configuration,
    AuthenticationError
]